from fastapi import APIRouter, FastAPI, Request
from fastapi.openapi.docs import get_redoc_html
from fastapi.responses import HTMLResponse


def build_scalar_docs(app: FastAPI) -> APIRouter:
    # pylint: disable=import-outside-toplevel
    from scalar_fastapi import get_scalar_api_reference

    router = APIRouter()

    @router.get("/scalar", include_in_schema=False)
    async def scalar_docs(request: Request):
        root_path = request.scope.get("root_path", "")
        openapi_url = f"{root_path}{app.openapi_url}"
        return get_scalar_api_reference(
            openapi_url=openapi_url,
        )

    return router


# https://cdn.jsdelivr.net/npm/redoc/bundles/redoc.standalone.js
def build_redoc_docs(app, redoc_url):
    router = APIRouter()

    @router.get(redoc_url, include_in_schema=False)
    async def redoc(request: Request) -> HTMLResponse:
        root_path = request.scope.get("root_path", "")
        openapi_url = f"{root_path}{app.openapi_url}"
        return get_redoc_html(
            openapi_url=openapi_url,
            title=f"{request.app.state.schema.title} - ReDoc",
            redoc_js_url="https://cdn.jsdelivr.net/npm/redoc/bundles/redoc.standalone.js",
        )

    return router



from pathlib import PurePosixPath

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import HTMLResponse
from fastapi.templating import Jinja2Templates
from jinja2 import Environment, PackageLoader, select_autoescape

from brilliance_admin.schema import AdminSchema

router = APIRouter()

templates = Jinja2Templates(
    env=Environment(
        loader=PackageLoader("brilliance_admin", "templates"),
        autoescape=select_autoescape(["html", "xml"]),
    )
)

# Всё, что не должно попадать в SPA (можете расширять список)
EXACT_BLOCK = {"/openapi.json"}
PREFIX_BLOCK = ("/docs", "/redoc", "/scalar", "/static")


@router.get('/{rest_of_path:path}', response_class=HTMLResponse, include_in_schema=False)
async def admin_index(request: Request, rest_of_path: str):
    '''
    The request responds with a pre-rendered SPA served as an HTML page.
    '''

    path = PurePosixPath('/' + rest_of_path)

    if '..' in path.parts:
        raise HTTPException(status_code=404)

    path_str = str(path)
    if path_str in EXACT_BLOCK or path_str.startswith(PREFIX_BLOCK):
        raise HTTPException(status_code=404)

    schema: AdminSchema = request.app.state.schema

    return templates.TemplateResponse(
        request=request,
        name='index.html',
        context=await schema.get_index_context_data(request),
    )



import abc

from pydantic import BaseModel
from pydantic.dataclasses import dataclass
from brilliance_admin.utils import DataclassBase


@dataclass
class UserABC(DataclassBase, abc.ABC):
    username: str


class AuthData(BaseModel):
    username: str
    password: str


class UserResult(BaseModel):
    username: str


class AuthResult(BaseModel):
    token: str
    user: UserResult


class AdminAuthentication(abc.ABC):
    async def login(self, data: AuthData, debug: bool = False) -> AuthResult:
        raise NotImplementedError('Login is not implemented')

    async def authenticate(self, headers: dict) -> UserABC:
        raise NotImplementedError('authenticate is not implemented')



from typing import Any, List

from pydantic.dataclasses import dataclass

from brilliance_admin.auth import UserABC
from brilliance_admin.exceptions import AdminAPIException, APIError, FieldError
from brilliance_admin.schema.category import FieldSchemaData
from brilliance_admin.schema.table.fields.base import TableField
from brilliance_admin.schema.table.table_models import AutocompleteData, Record
from brilliance_admin.translations import LanguageContext
from brilliance_admin.translations import TranslateText as _
from brilliance_admin.utils import DeserializeAction


def get_pk(obj):
    pk_cols = obj.__mapper__.primary_key
    if len(pk_cols) != 1:
        raise NotImplementedError('Composite primary key is not supported')
    return getattr(obj, pk_cols[0].key)


@dataclass
class SQLAlchemyRelatedField(TableField):
    _type: str = 'related'

    # Тип связи.
    # Откуда берётся:
    # - из SQLAlchemy relationship.uselist
    #   rel.uselist == True  -> список связанных объектов
    #   rel.uselist == False -> одиночная связь
    #
    # Зачем нужен:
    # - чтобы понимать, ожидать list или один объект
    # - влияет на логику update_related и serialize
    many: bool = False

    # Имя relationship-атрибута на модели.
    # Откуда берётся:
    # - из mapper.relationships: rel.key
    # - либо через поиск relationship по FK колонке (col.local_columns)
    #
    # Зачем нужен:
    # - для доступа к связи через ORM
    #   getattr(record, rel_name)
    # - для записи и чтения связанных объектов
    rel_name: str | None

    # Класс связанной SQLAlchemy-модели.
    # Откуда берётся:
    # - из relationship: rel.mapper.class_
    #
    # Зачем нужен:
    # - для загрузки связанных записей из БД
    #   session.get(target_model, pk)
    #   select(target_model).where(target_model.id.in_(...))
    target_model: Any | None = None

    # Работает только если many=True
    dual_list: bool = False

    def generate_schema(self, user: UserABC, field_slug, language_context: LanguageContext) -> FieldSchemaData:
        schema = super().generate_schema(user, field_slug, language_context)
        schema.many = self.many
        schema.rel_name = self.rel_name
        schema.dual_list = self.dual_list
        return schema

    def _get_target_model(self, model, field_slug):
        # pylint: disable=import-outside-toplevel
        from sqlalchemy import inspect

        mapper = inspect(model).mapper
        attr = mapper.attrs.get(field_slug)
        if attr is None:
            msg = f'Field "{field_slug}" is not found on model "{model}"'
            raise AttributeError(msg)

        # RelationshipProperty
        if hasattr(attr, 'mapper'):
            return attr.mapper.class_

        # ColumnProperty (FK column). Try to resolve from foreign key target table.
        col = getattr(model, field_slug).property.columns[0]
        if not col.foreign_keys:
            msg = f'Field "{field_slug}" is not a relationship and not a FK column'
            raise AttributeError(msg)

        fk = next(iter(col.foreign_keys))
        target_table = fk.column.table

        # Find a mapped class that uses this table in the same registry
        for m in mapper.registry.mappers:
            if getattr(m, 'local_table', None) is target_table:
                return m.class_

        msg = f'Cannot resolve target model for FK "{field_slug}"'
        raise AttributeError(msg)

    async def autocomplete(self, model, data: AutocompleteData, user, *, extra: dict | None = None) -> List[Record]:
        # pylint: disable=import-outside-toplevel
        from sqlalchemy import select

        if extra is None or extra.get('db_async_session') is None:
            msg = f'SQLAlchemyRelatedField.autocomplete {type(self).__name__} requires extra["db_async_session"] (AsyncSession)'
            raise AttributeError(msg)

        db_async_session = extra['db_async_session']

        results = []

        target_model = self._get_target_model(model, data.field_slug)
        limit = min(150, data.limit)
        stmt = select(target_model).limit(limit)

        pk = get_pk(target_model)
        python_pk_type = pk.property.columns[0].type.python_type

        if data.search_string:
            try:
                value = python_pk_type(data.search_string)
            except (ValueError, TypeError):
                # Search string cannot be cast to primary key type, skip id filter
                value = None

            stmt = stmt.where(pk == value)

        # Add already selected choices
        if data.existed_choices:
            existed_choices = [i['key'] for i in data.existed_choices if 'key' in i]

            values = []
            for value in existed_choices:
                try:
                    values.append(python_pk_type(value))
                except (ValueError, TypeError) as e:
                    msg = f'Invalid existed_choices value "{value}" for pk {pk} python_pk_type:{python_pk_type.__name__}'
                    raise AdminAPIException(APIError(message=msg), status_code=500) from e

            stmt = stmt.where(pk.in_(values))

        async with db_async_session() as session:
            records = (await session.execute(stmt)).scalars().all()

        for record in records:
            results.append(Record(key=getattr(record, pk.key), title=str(record)))

        return results

    async def serialize(self, value, extra: dict, *args, **kwargs) -> Any:
        """
        Сериализация related-поля.

        Входные данные:
        - value всегда scalar (None или int)
        - ORM-объект доступен через extra["record"]
        """
        if not value:
            return

        record = extra.get('record')
        if record is None:
            raise FieldError(f'Missing record in serialize context in value: {value}')

        related = getattr(record, self.rel_name, None)

        if self.many:
            if related is None:
                raise FieldError(f'Many Related field "{self.rel_name}" is missing on record "{record}"')
            return [{'key': get_pk(obj), 'title': str(obj)} for obj in related]

        if related is None:
            return None

        return {'key': get_pk(related), 'title': str(related)}

    async def deserialize(self, value, action: DeserializeAction, extra: dict, *args, **kwargs) -> Any:
        value = await super().deserialize(value, action, extra, *args, **kwargs)
        if not value:
            return None

        if isinstance(value, list):
            result = []
            for i in value:
                i = i.get('key')
                if not isinstance(i, (int, str)):
                    raise FieldError(f'Value "{i}" is not supported for related field')
                result.append(i)
            return result

        result = None
        if isinstance(value, dict) and 'key' in value:
            result = value['key']

        if isinstance(value, (int, str)):
            result = value

        if not isinstance(result, (int, str)):
            raise FieldError(f'Value "{result}" is not supported for related field')

        return result

    async def update_related(self, record, field_slug, value, session):
        """
        Обновление SQLAlchemy relationship.

        Предположения:
        - self.rel_name всегда имя relationship
        - self.target_model задан
        - self.many отражает тип связи
        """

        # pylint: disable=import-outside-toplevel

        if value is None:
            return

        # При CREATE объект должен быть в session до работы с relationship
        if record not in session:
            session.add(record)

        rel_attr = self.rel_name

        if self.many:
            assert isinstance(value, list)

            if not value:
                setattr(record, rel_attr, [])
                return

            result = []
            for i in value:
                obj = await session.get(self.target_model, i)
                if obj is None:
                    msg = _('related_not_found') % {
                        'model': self.target_model.__name__,
                        'pk': i,
                        'field_slug': field_slug,
                    }
                    raise AdminAPIException(
                        APIError(message=msg, code='related_not_found'),
                        status_code=400,
                    )
                result.append(obj)

            # getattr(record, rel_attr).clear()
            getattr(record, rel_attr).extend(list(result))
            return

        obj = await session.get(self.target_model, value)
        setattr(record, rel_attr, obj)

    async def apply_filter(self, stmt, value, model, column):
        # pylint: disable=import-outside-toplevel
        from sqlalchemy import inspect

        if value is None:
            return stmt

        rel = getattr(model, self.rel_name)
        pk_col = inspect(self.target_model).primary_key[0]

        # many=False: FK (many-to-one)
        if not self.many:
            if not isinstance(value, int):
                raise FieldError(f'Expected int for filter {self.rel_name}')
            return stmt.where(rel.has(pk_col == value))

        # many=True: one-to-many / many-to-many
        if not isinstance(value, list):
            raise FieldError(f'Expected list[int] for filter {self.rel_name}')
        return stmt.where(rel.any(pk_col.in_(value)))



import importlib.metadata
import json
from importlib import resources
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from pydantic import Field
from pydantic.dataclasses import dataclass

from brilliance_admin.auth import UserABC
from brilliance_admin.docs import build_redoc_docs, build_scalar_docs
from brilliance_admin.schema.category import BaseCategory, CategorySchemaData
from brilliance_admin.translations import LanguageContext, LanguageManager
from brilliance_admin.utils import DataclassBase, SupportsStr

DEFAULT_LANGUAGES = {
    'ru': 'Russian',
    'en': 'English',
}


@dataclass
class AdminSchemaData(DataclassBase):
    profile: UserABC | Any
    categories: Dict[str, CategorySchemaData] = Field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.profile, UserABC):
            self.profile = UserABC(username=self.profile.username)


# pylint: disable=too-many-instance-attributes
@dataclass
class AdminSettingsData(DataclassBase):
    title: SupportsStr
    description: SupportsStr | None
    login_greetings_message: SupportsStr | None
    navbar_density: str
    languages: Dict[str, str] | None
    main_page: str | None = None


@dataclass
class AdminIndexContextData(DataclassBase):
    title: str
    favicon_image: str | None
    settings_json: str


@dataclass
class AdminSchema:
    categories: List[BaseCategory]
    auth: Any

    main_page: str | None = None

    title: SupportsStr | None = 'Admin'
    description: SupportsStr | None = None
    login_greetings_message: SupportsStr | None = None

    logo_image: str | None = None
    favicon_image: str | None = None

    navbar_density: str = 'default'

    backend_prefix: str | None = None
    static_prefix: str | None = None

    language_manager: LanguageManager | None = None

    debug: bool = False

    def __post_init__(self):
        for category in self.categories:
            if not issubclass(category.__class__, BaseCategory):
                raise TypeError(f'Root category "{category}" is not instance of BaseCategory subclass')

        if not self.language_manager:
            self.language_manager = LanguageManager(DEFAULT_LANGUAGES)

    def get_language_context(self, language_slug: str | None) -> LanguageContext:
        return LanguageContext(language_slug, language_manager=self.language_manager)

    def generate_schema(self, user: UserABC, language_slug: str | None) -> AdminSchemaData:
        language_context: LanguageContext = self.get_language_context(language_slug)

        result = AdminSchemaData(profile=user)

        for category in self.categories:
            if not category.slug:
                msg = f'Category {type(category).__name__}.slug is empty'
                raise AttributeError(msg)

            try:
                result.categories[category.slug] = category.generate_schema(user, language_context).to_dict(keep_none=False)
            except Exception as e:
                msg = f'Root category "{category.slug}" generate_schema error: {e}'
                raise Exception(msg) from e

        return result

    def get_group(self, group_slug: str) -> Optional[BaseCategory]:
        for category in self.categories:
            if category.slug == group_slug:
                return category

        return None

    async def get_settings(self, request: Request) -> AdminSettingsData:
        language_slug = request.headers.get('Accept-Language')
        language_context: LanguageContext = self.get_language_context(language_slug)

        languages = None
        if self.language_manager.languages:
            languages = {}
            for k, v in self.language_manager.languages.items():
                languages[k] = v

        return AdminSettingsData(
            title=self.title,
            main_page=self.main_page,
            description=self.description,
            login_greetings_message=self.login_greetings_message,
            navbar_density=self.navbar_density,
            languages=languages,
        )

    def generate_app(
            self,
            debug=False,
            allow_cors=True,

            include_scalar=False,
            include_docs=False,
            include_redoc=False,
    ) -> FastAPI:
        self.debug = debug

        # pylint: disable=unused-variable
        language_context = self.get_language_context(language_slug=None)

        app = FastAPI(
            title=language_context.get_text(self.title),
            description=language_context.get_text(self.description),
            debug=debug,
            docs_url='/docs' if include_docs else None,
            redoc_url=None,
        )

        if allow_cors:
            app.add_middleware(
                CORSMiddleware,
                allow_origins=["*"],
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"]
            )

        static_dir = resources.files("brilliance_admin").joinpath("static")
        app.mount("/static", StaticFiles(directory=str(static_dir)), name="static")

        app.state.schema = self

        if include_scalar:
            app.include_router(build_scalar_docs(app))

        if include_redoc:
            app.include_router(build_redoc_docs(app, redoc_url='/redoc'))

        # pylint: disable=import-outside-toplevel
        from brilliance_admin.api.routers import brilliance_admin_router
        app.include_router(brilliance_admin_router)

        return app

    async def get_index_context_data(self, request: Request) -> dict:
        language_context = self.get_language_context(language_slug=None)
        context = {'language_context': language_context}

        backend_prefix = self.backend_prefix
        if not backend_prefix:
            backend_prefix = urljoin(str(request.base_url), '/admin/')

        static_prefix = self.static_prefix
        if not static_prefix:
            static_prefix = urljoin(str(request.base_url), '/admin/static/')

        logo_image = self.logo_image
        if logo_image and logo_image.startswith('/'):
            logo_image = urljoin(str(request.base_url), logo_image)

        settings_json = {
            'backend_prefix': backend_prefix,
            'static_prefix': static_prefix,
            'version': importlib.metadata.version('brilliance-admin'),
            'api_timeout_ms': 1000 * 5,
            'logo_image': logo_image,
        }
        data = AdminIndexContextData(
            title=str(self.title),
            favicon_image=self.favicon_image,
            settings_json=json.dumps(settings_json),
        )
        return data.model_dump(mode='json', context=context)



from brilliance_admin import auth, schema
from brilliance_admin.exceptions import AdminAPIException, APIError, FieldError
from brilliance_admin.integrations.sqlalchemy.fields_schema import SQLAlchemyFieldsSchema
from brilliance_admin.schema.admin_schema import AdminSchema
from brilliance_admin.translations import LanguageContext
from brilliance_admin.translations import TranslateText as _
from brilliance_admin.utils import get_logger

logger = get_logger()


class SQLAlchemyAdminListMixin:
    table_schema: SQLAlchemyFieldsSchema
    table_filters: SQLAlchemyFieldsSchema | None

    def apply_ordering(self, stmt, list_data):
        # pylint: disable=import-outside-toplevel
        from sqlalchemy import asc, desc
        from sqlalchemy.orm import InstrumentedAttribute

        ordering = list_data.ordering or self.default_ordering

        if not ordering:
            return stmt

        direction = asc

        if ordering.startswith("-"):
            ordering = ordering[1:]
            direction = desc

        if list_data.ordering and ordering not in self.ordering_fields:
            msg = f'Ordering "{ordering}" is not allowed; available options: {self.ordering_fields} default_ordering: {self.default_ordering}'
            raise FieldError(message=msg, field_slug='ordering')

        column = getattr(self.model, ordering, None)
        if not isinstance(column, InstrumentedAttribute):
            msg = f'{type(self).__name__} ordering field "{ordering}" not found in model {self.model}'
            raise FieldError(message=msg, field_slug='ordering')

        return stmt.order_by(direction(column))

    def apply_search(self, stmt, list_data: schema.ListData):
        # pylint: disable=import-outside-toplevel
        from sqlalchemy import String, cast, or_
        from sqlalchemy.orm import InstrumentedAttribute

        if not self.search_fields or not list_data.search:
            return stmt

        search = list_data.search
        conditions = []

        for field_slug in self.search_fields:
            column = getattr(self.model, field_slug, None)
            if not isinstance(column, InstrumentedAttribute):
                msg = f'{type(self).__name__} filter "{field_slug}" not found as field inside model {self.model}'
                raise AttributeError(msg)

            conditions.append(cast(column, String).ilike(search))

        if conditions:
            stmt = stmt.where(or_(*conditions))

        return stmt

    async def apply_filters(self, stmt, list_data: schema.ListData):
        if not self.table_filters or not list_data.filters:
            return stmt

        if not issubclass(type(self.table_filters), SQLAlchemyFieldsSchema):
            msg = f'{type(self).__name__}.table_filters {type(self.table_filters)} must be SQLAlchemyFieldsSchema subclass'
            raise AttributeError(msg)

        return await self.table_filters.apply_filters(stmt, list_data.filters)

    def apply_pagination(self, stmt, list_data: schema.ListData):
        page = max(1, list_data.page or 1)
        limit = min(150, max(1, list_data.limit or 25))

        offset = (page - 1) * limit

        return stmt.limit(limit).offset(offset)

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    async def get_list(
        self,
        list_data: schema.ListData,
        user: auth.UserABC,
        language_context: LanguageContext,
        admin_schema: AdminSchema,
    ) -> schema.TableListResult:
        # pylint: disable=import-outside-toplevel
        from sqlalchemy import exc, func, select

        try:
            stmt = self.get_queryset()
            stmt = await self.apply_filters(stmt, list_data)
            stmt = self.apply_search(stmt, list_data)

            count_stmt = select(func.count()).select_from(stmt.subquery())
            stmt = self.apply_pagination(stmt, list_data)
            stmt = self.apply_ordering(stmt, list_data)

        except FieldError as e:
            logger.exception(
                'SQLAlchemy %s list filters for %s field error: %s',
                type(self).__name__, self.model.__name__, e,
                extra={
                    'list_data': list_data,
                }
            )
            msg = _('errors.filter_error') % {'error': e.message}
            raise AdminAPIException(APIError(message=msg, code='filters_exception'), status_code=500) from e

        except Exception as e:
            logger.exception(
                'SQLAlchemy %s list filters for %s error: %s',
                type(self).__name__, self.model.__name__, e,
                extra={
                    'list_data': list_data,
                }
            )
            msg = _('errors.filters_exception') % {
                'error': str(e) if admin_schema.debug else type(e).__name__,
            }
            raise AdminAPIException(APIError(message=msg, code='filters_exception'), status_code=500) from e

        try:
            async with self.db_async_session() as session:
                total_count = await session.scalar(count_stmt)
                records = (await session.execute(stmt)).scalars().all()

        except ConnectionRefusedError as e:
            logger.exception(
                'SQLAlchemy %s get_list db error: %s',
                type(self).__name__, e,
                extra={
                    'list_data': list_data,
                }
            )
            msg = _('errors.connection_refused_error') % {'error': str(e)}
            raise AdminAPIException(
                APIError(message=msg, code='connection_refused_error'),
                status_code=500,
            ) from e

        except (exc.IntegrityError, exc.StatementError) as e:
            logger.exception(
                'SQLAlchemy %s get_list db error: %s',
                type(self).__name__, e,
                extra={
                    'list_data': list_data,
                }
            )
            orig = e.orig
            message = orig.args[0] if orig.args else type(orig).__name__
            raise AdminAPIException(
                APIError(message=message, code='db_exception'), status_code=500,
            ) from e

        except Exception as e:
            logger.exception(
                'SQLAlchemy %s get_list db error: %s',
                type(self).__name__, e,
                extra={
                    'list_data': list_data,
                }
            )
            msg = _('errors.db_error_list') % {
                'error_type': str(e) if admin_schema.debug else type(e).__name__,
            }
            raise AdminAPIException(
                APIError(message=msg, code='db_error_list'), status_code=500,
            ) from e

        try:
            data = []
            for record in records:
                line = await self.table_schema.serialize(
                    record,
                    extra={"record": record, "user": user},
                )
                data.append(line)

        except FieldError as e:
            logger.exception(
                'SQLAlchemy %s list %s serialize field error: %s',
                type(self).__name__, self.model.__name__, e,
            )
            msg = _('serialize_error.field_error') % {
                'error': e.message,
                'field_slug': e.field_slug,
            }
            raise AdminAPIException(APIError(message=msg, code='field_error'), status_code=500) from e

        except Exception as e:
            logger.exception(
                'SQLAlchemy %s list %s serialize error: %s',
                type(self).__name__, self.model.__name__, e,
            )
            msg = _('serialize_error.unexpected_error') % {
                'error': str(e) if admin_schema.debug else type(e).__name__,
            }
            raise AdminAPIException(APIError(message=msg, code='unexpected_error'), status_code=500) from e

        return schema.TableListResult(data=data, total_count=int(total_count or 0))



from unittest import mock

import pytest

from brilliance_admin.auth import UserABC
from example.sections.payments import PaymentsAdmin

category_schema_data = {
    'dashboard_info': None,
    'icon': 'mdi-credit-card-outline',
    'description': 'Статичные данные',
    'link': None,
    'categories': {},
    'table_info': {
        'actions': {
            'action_with_exception': {
                'allow_empty_selection': True,
                'base_color': None,
                'confirmation_text': None,
                'description': None,
                'form_schema': None,
                'icon': None,
                'title': 'Действие с ошибкой',
                'variant': None,
            },
            'create_payment': {
                'allow_empty_selection': True,
                'base_color': None,
                'confirmation_text': None,
                'description': 'Создать платеж и отправить его на обработку в платежную '
                'систему.',
                'form_schema': {
                    'fields': {
                        'amount': {
                            'header': {},
                            'label': 'Сумма',
                            'read_only': False,
                            'required': False,
                            'type': 'integer',
                        },
                        'is_throw_error': {
                            'header': {},
                            'label': 'Выбросить ошибку?',
                            'read_only': False,
                            'required': False,
                            'type': 'boolean',
                        },
                    },
                    'list_display': [
                        'amount',
                        'is_throw_error',
                    ],
                },
                'icon': None,
                'title': 'Создать платеж',
                'variant': None,
            },
            'delete': {
                'allow_empty_selection': False,
                'base_color': 'red-lighten-2',
                'confirmation_text': 'Вы уверены, что хотите удалить данные записи?\n'
                'Данное действие нельзя отменить.',
                'description': None,
                'form_schema': None,
                'icon': None,
                'title': 'Удалить',
                'variant': 'outlined',
            },
        },
        'can_create': False,
        'can_retrieve': True,
        'can_update': False,
        'ordering_fields': [
            'id',
        ],
        'default_ordering': None,
        'pk_name': 'id',
        'search_enabled': True,
        'search_help': mock.ANY,
        'table_filters': {
            'fields': {
                'created_at': {
                    'header': {},
                    'label': 'Время создания',
                    'range': True,
                    'read_only': False,
                    'required': False,
                    'type': 'datetime',
                    'include_date': True,
                    'include_time': True,
                },
                'id': {
                    'header': {},
                    'label': 'ID',
                    'read_only': False,
                    'required': False,
                    'type': 'integer',
                },
            },
            'list_display': [
                'id',
                'created_at',
            ],
        },
        'table_schema': {
            'fields': {
                'amount': {
                    'header': {},
                    'label': 'Сумма',
                    'read_only': True,
                    'required': False,
                    'type': 'integer',
                },
                'created_at': {
                    'header': {},
                    'label': 'Время создания',
                    'read_only': True,
                    'required': False,
                    'type': 'datetime',
                    'include_date': True,
                    'include_time': True,
                },
                'description': {
                    'header': {},
                    'label': 'Описание',
                    'read_only': False,
                    'required': False,
                    'type': 'string',
                },
                'endpoint': {
                    'header': {},
                    'label': 'Эндпоинт',
                    'read_only': False,
                    'required': False,
                    'type': 'string',
                },
                'get_provider_registry': {
                    'header': {},
                    'label': 'Реестр проверен',
                    'read_only': True,
                    'required': False,
                    'type': 'boolean',
                },
                'get_provider_registry_info': {
                    'header': {},
                    'label': 'Информация по реестру провайдера',
                    'read_only': True,
                    'required': False,
                    'type': 'boolean',
                },
                'id': {
                    'header': {},
                    'label': 'ID',
                    'read_only': True,
                    'required': False,
                    'type': 'integer',
                },
                'other_field': {
                    'header': {},
                    'label': 'Other Field',
                    'read_only': True,
                    'required': False,
                    'type': 'string',
                },
                'whitelist_ips': {
                    'header': {},
                    'label': 'Белый список IP',
                    'read_only': False,
                    'required': False,
                    'type': 'array',
                },
            },
            'list_display': [
                'id',
                'amount',
                'endpoint',
                'description',
                'created_at',
                'get_provider_registry',
                'get_provider_registry_info',
            ],
        },
    },
    'title': 'Платежи',
    'type': 'table',
 }


@pytest.mark.asyncio
async def test_generate_category_schema(language_context):
    category = PaymentsAdmin()
    new_schema = category.generate_schema(UserABC(username="test"), language_context)
    assert new_schema.model_dump() == category_schema_data, new_schema.model_dump()



# pylint: disable=wildcard-import, unused-wildcard-import, unused-import, too-many-ancestors
# flake8: noqa: F405
from .base import SQLAlchemyAdminBase
from .create import SQLAlchemyAdminCreate
from .delete import SQLAlchemyDeleteAction
from .list import SQLAlchemyAdminListMixin
from .retrieve import SQLAlchemyAdminRetrieveMixin
from .update import SQLAlchemyAdminUpdate


class SQLAlchemyAdmin(
        SQLAlchemyAdminUpdate,
        SQLAlchemyAdminCreate,
        SQLAlchemyDeleteAction,
        SQLAlchemyAdminListMixin,
        SQLAlchemyAdminRetrieveMixin,
        SQLAlchemyAdminBase,
):
    pass



import abc
from typing import Any, ClassVar, Dict, List

from pydantic import Field
from pydantic.dataclasses import dataclass
from pydantic_core import core_schema
from structlog import get_logger

from brilliance_admin.auth import UserABC
from brilliance_admin.translations import LanguageContext
from brilliance_admin.utils import DataclassBase, KwargsInitMixin, SupportsStr, humanize_field_name

logger = get_logger()


# pylint: disable=too-many-instance-attributes
@dataclass
class FieldSchemaData(DataclassBase):
    type: str = None

    label: str | None = None
    help_text: str | None = None

    # Table header parameters
    header: dict = Field(default_factory=dict)

    read_only: bool = False
    default: Any | None = None
    required: bool = False

    max_length: int | None = None
    min_length: int | None = None

    choices: List[dict] | None = None

    variant: str | None = None
    size: str | None = None

    preview_max_height: int | None = None
    preview_max_width: int | None = None

    # StringField
    multilined: bool | None = None
    ckeditor: bool | None = None
    tinymce: bool | None = None

    # ArrayField
    array_type: str | None = None

    # SQLAlchemyRelatedField
    many: bool | None = None
    rel_name: str | None = None
    dual_list: bool | None = None

    # IntegerField
    inputmode: str | None = None
    precision: int | None = None
    scale: int | None = None

    # DateTimeField
    range: bool | None = None
    include_date: bool | None = None
    include_time: bool | None = None


@dataclass
class FieldsSchemaData(DataclassBase):
    fields: Dict[str, dict] = Field(default_factory=dict)
    list_display: List[str] = Field(default_factory=list)


# pylint: disable=too-many-instance-attributes
@dataclass
class TableInfoSchemaData(DataclassBase):
    table_schema: FieldsSchemaData

    search_enabled: bool = Field(default=False)
    search_help: str | None = Field(default=None)

    pk_name: str | None = Field(default=None)
    can_retrieve: bool = Field(default=False)

    can_create: bool = Field(default=False)
    can_update: bool = Field(default=False)

    table_filters: FieldsSchemaData | None = Field(default=None)

    ordering_fields: List[str] = Field(default_factory=list)
    default_ordering: str | None = None

    actions: Dict[str, dict] | None = Field(default_factory=dict)

    def __repr__(self):
        return f'<TableInfoSchemaData id={id(self)}>'


@dataclass
class DashboardInfoSchemaData(DataclassBase):
    search_enabled: bool
    search_help: str | None

    table_filters: FieldsSchemaData | None = None


@dataclass
class CategorySchemaData(DataclassBase):
    title: str | None
    description: str | None
    icon: str | None
    type: str

    categories: dict = Field(default_factory=dict)

    table_info: TableInfoSchemaData | None = None
    dashboard_info: DashboardInfoSchemaData | None = None

    link: str | None = None

    def __repr__(self):
        return f'<CategorySchemaData type={self.type} "{self.title}">'


class BaseCategory(KwargsInitMixin, abc.ABC):
    slug: str
    title: SupportsStr | None = None
    description: SupportsStr | None = None

    # https://pictogrammers.com/library/mdi/
    icon: str | None = None

    _type_slug: ClassVar[str]

    def generate_schema(self, user: UserABC, language_context: LanguageContext) -> CategorySchemaData:
        type_slug = getattr(type(self), '_type_slug', None)
        if not type_slug:
            msg = f'{type(self).__name__}._type_slug must be set!'
            raise AttributeError(msg)

        result = CategorySchemaData(
            title=language_context.get_text(self.title) or humanize_field_name(self.slug),
            description=language_context.get_text(self.description),
            icon=self.icon,
            type=type_slug,
        )
        return result

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if cls is BaseCategory:
            return

        if not issubclass(cls, BaseCategory):
            raise TypeError(f'{cls.__name__} must inherit from Category')

    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: Any) -> core_schema.CoreSchema:
        def validate(v: Any) -> "BaseCategory":
            if isinstance(v, cls):
                return v
            raise TypeError(f"Expected {cls.__name__} instance, recieved: {type(v)} {v}")

        return core_schema.no_info_plain_validator_function(
            validate,
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: repr(v),
                info_arg=False,
                return_schema=core_schema.str_schema(),
            ),
        )


class CategoryLink(BaseCategory):
    _type_slug: str = 'link'

    link: str

    def generate_schema(self, user: UserABC, language_context: LanguageContext) -> CategorySchemaData:
        result = super().generate_schema(user, language_context)
        result.link = self.link
        return result


class CategoryGroup(BaseCategory):
    _type_slug: str = 'group'

    subcategories: list = Field(default_factory=list)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for category in self.subcategories:
            if not isinstance(category, BaseCategory):
                raise TypeError(f'Category "{category}" is not instance of BaseCategory subclass')

    def generate_schema(self, user: UserABC, language_context: LanguageContext) -> CategorySchemaData:
        result = super().generate_schema(user, language_context)

        for category in self.subcategories:

            if not category.slug:
                msg = f'Category {type(category).__name__}.slug is empty'
                raise AttributeError(msg)

            if category.slug in result.categories:
                exists = result.categories[category.slug]
                msg = f'Category {type(category).__name__}.slug "{self.slug}" already registered by "{exists.title}"'
                raise KeyError(msg)

            try:
                result.categories[category.slug] = category.generate_schema(user, language_context)
            except Exception as e:
                msg = f'Category "{category.slug}" {type(category)} generate_schema error: {e}'
                raise Exception(msg) from e

        return result

    def get_category(self, category_slug: str):
        for category in self.subcategories:
            if category.slug == category_slug:
                return category

        return None



from brilliance_admin.exceptions import APIError, AdminAPIException
from brilliance_admin.schema.table.admin_action import ActionData, ActionMessage, ActionResult, admin_action
from brilliance_admin.translations import TranslateText as _


class SQLAlchemyDeleteAction:
    has_delete: bool = True

    @admin_action(
        title=_('delete'),
        confirmation_text=_('delete_confirmation_text'),
        base_color='red-lighten-2',
        variant='outlined',
    )
    async def delete(self, *args, action_data: ActionData, **kwargs):
        if not self.has_delete:
            raise AdminAPIException(APIError(message=_('errors.method_not_allowed')), status_code=500)
        return ActionResult(message=ActionMessage(_('deleted_successfully')))



import functools
import inspect
from typing import Any

from pydantic.dataclasses import dataclass

from brilliance_admin.exceptions import AdminAPIException, APIError
from brilliance_admin.schema.table.fields.base import TableField
from brilliance_admin.utils import get_logger

logger = get_logger()


def function_field(**kwargs):
    '''
    The same as decaring:
    field = FunctionField(fn=attribute)

    but available directly and converted after to the FunctionField
    '''
    def wrapper(func):
        func.__function_field__ = True

        field_type = kwargs.pop('type', None)
        if field_type:
            # pylint: disable=protected-access
            kwargs['_type'] = field_type._type

        func.__kwargs__ = kwargs

        @functools.wraps(func)
        async def wrapped(*args, **kwargs):
            return await func(*args, **kwargs)

        return wrapped

    return wrapper


@dataclass
class FunctionField(TableField):
    _type: str = 'function_field'
    read_only = True

    fn: Any = None

    def __post_init__(self):
        if not inspect.iscoroutinefunction(self.fn):
            msg = f'{type(self).__name__}.fn {self.fn} must be coroutine function'
            raise AttributeError(msg)

    async def serialize(self, value, extra: dict, *args, **kwargs) -> Any:
        try:
            return await self.fn(**extra)
        except Exception as e:
            logger.exception(
                'Function field %s label=%s error from function="%s": %s',
                type(self).__name__,
                self.label,
                self.fn,
                e,
            )
            raise AdminAPIException(
                APIError(message=f'Error: {e}', code='function_field_error'), status_code=400,
            ) from e



from fastapi import HTTPException

from brilliance_admin.auth import AdminAuthentication


async def get_user(request):
    auth: AdminAuthentication = request.app.state.schema.auth
    user = await auth.authenticate(request.headers)
    return user


async def get_category(request, group: str, category: str, check_type=None):
    user = await get_user(request)

    schema_group = request.app.state.schema.get_group(group)
    if not schema_group:
        raise HTTPException(status_code=404, detail="Group not found")

    schema_category = schema_group.get_category(category)
    if not schema_category:
        raise HTTPException(status_code=404, detail="Category not found")

    if check_type:
        schema_category, user = await get_category(request, group, category)
        if not issubclass(schema_category.__class__, check_type):
            raise HTTPException(status_code=404, detail=f"Category {group}.{category} is not a {check_type.__name__}")

    return schema_category, user
